@@ -1,6 +1,6 @@
 """FASTGenomics base library"""
 
-__version__ = "3.3.1"
+__version__ = "3.4.0"
 
 from .fg import (
     FASTGenomicsClient,
@@ -9,3 +9,7 @@ from .fg import (
     ToolConfiguration,
     run_zip,
 )
+
+from .FASTGenomicsPlatformUpdater import (
+    FASTGenomicsPlatformUpdater
+)
