@@ -22,9 +22,10 @@ from subprocess import CompletedProcess
 from typing import List, Union
 from urllib.parse import urlparse
 from . import __version__
+from urllib.parse import urlparse
 
 import requests
-from requests.adapters import Retry
+from requests.adapters import Response, Retry
 from requests.exceptions import HTTPError
 from requests.sessions import HTTPAdapter
 from tqdm import tqdm
@@ -44,6 +45,10 @@ def get_homedir() -> str:
     return path.normpath(path.expanduser("~/.fgcli"))
 
 
+def is_absolute(url):
+    return bool(urlparse(url).netloc)
+
+
 def run_zip(
     files_or_directory: List[str], zip_filename, zip_password: str
 ) -> str:
@@ -84,10 +89,11 @@ class FASTGenomicsAccount:
 
     def __init__(self):
         """ constructs a fastgenomics account """
-        self.fastgenomics_url = ""
-        self.fastgenomics_user = ""
-        self.fastgenomics_passphrase = ""
-        self.fastgenomics_login_method = ""
+        self.identityserver_url: str = ""
+        self.fastgenomics_url: str = ""
+        self.fastgenomics_user: str = ""
+        self.fastgenomics_passphrase: str = ""
+        self.fastgenomics_login_method: str = ""
 
     def to_json(self):
         """ converts the account to json """
@@ -618,6 +624,37 @@ class FASTGenomicsClient:
         """ get the version """
         return {"fg.py": f"{FG_CLI_VERSION}"}
 
+    def _get_identityserver_url(self, url: str) -> str:
+        """ get the identityserver url from webclient settings """
+        try:
+            headers = {
+                "Content-type": "application/json",
+                "Accept": "application/json",
+                "User-Agent": user_agent,
+            }
+            if not url.endswith("/"):
+                url += "/"
+            url += "webclient/api/settings"
+            response = self.session.get(
+                url, headers=headers
+            )
+            self._handle_response(response)
+            jsonResponse = response.json()
+
+            ids_url = jsonResponse.get("idsUrl")
+            if not ids_url.endswith("/"):
+                ids_url = ids_url + "/"
+            return ids_url
+
+        except HTTPError as http_err:
+            raise Exception(
+                f"Failed to get settings from '{url}'': {http_err}"
+            )
+        except Exception as err:
+            raise Exception(
+                f"Failed to get settings from '{url}'': {err}"
+            )
+
     def _get_bearer_token(self) -> str:
         """ get a bearer token """
         account = self._get_fastgenomics_account()
@@ -633,7 +670,7 @@ class FASTGenomicsClient:
             }
 
             if account.fastgenomics_login_method == "pat":
-                url = self._get_url("ids/api/v1/token/pat")
+                url = self.get_identityserver_url("api/v1/token/pat")
 
                 request_data = {
                     "email": account.fastgenomics_user,
@@ -648,7 +685,7 @@ class FASTGenomicsClient:
                 self._handle_response(response)
                 jsonResponse = response.json()
             else:
-                url = self._get_url("ids/api/v1/token/resourceowner")
+                url = self.get_identityserver_url("api/v1/token/resourceowner")
 
                 request_data = {
                     "email": account.fastgenomics_user,
@@ -749,15 +786,14 @@ class FASTGenomicsClient:
         return account
 
     def login(self, user: str, passphrase: str, login_method: str, url: str):
-
-        """login to FASTGenomics and store credentials in users profile
+        """login to the platform and store credentials in users profile
 
         Parameters:
         -----------
         user         : the FASTGenomics user name
         passphrase   : if login_method is "pat" the personal access token (pat), if login_method is "bearer" the bearer token else the password of the user
         login_method : pat | password | bearer
-        url          : the url of the plattform
+        url          : the url of the platform
         """
         directory = FASTGenomicsClient._get_cli_profile_path()
         if not path.exists(directory):
@@ -785,6 +821,8 @@ class FASTGenomicsClient:
         account.fastgenomics_user = user
         account.fastgenomics_passphrase = passphrase
         account.fastgenomics_login_method = login_method
+        ids_url = self._get_identityserver_url(url)
+        account.identityserver_url = ids_url
 
         content = account.to_json()
         logger.info(f"Writing profile '{filename}'")
@@ -801,30 +839,214 @@ class FASTGenomicsClient:
                 raise Exception(f"Log in failed!: {err}")
 
     def logout(self):
-        """ logout from FASTGenomics and remove credentials file """
+        """ logout from the platform and remove credentials file """
         filename = FASTGenomicsClient.__get_credentials_filename()
         if path.exists(filename):
             remove(filename)
         self.fastgenomics_account = None
 
     def is_logged_in(self) -> bool:
+        """ is the user logged in ? """
         try:
             self._get_fastgenomics_account()
         except:
             return False
         return True
 
-    def _get_url(self, relative_url: str) -> str:
-        """ get a absolute FASTGenomics url """
+    def get_platform_url(self, relative_url: str) -> str:
+        """ get the absolute platform url """
         if relative_url.startswith("/"):
             relative_url = relative_url[1:]
         return self._get_fastgenomics_account().fastgenomics_url + relative_url
 
+    def get_backend_url(self, relative_url: str) -> str:
+        """ get the absolute backend url """
+        if relative_url.startswith("/"):
+            relative_url = relative_url[1:]
+        return self._get_fastgenomics_account().fastgenomics_url + "webclient/" + relative_url
+
+    def get_identityserver_url(self, relative_url: str) -> str:
+        """ get the absolute identityserver url """
+        if relative_url.startswith("/"):
+            relative_url = relative_url[1:]
+
+        url = self._get_fastgenomics_account().identityserver_url
+        if url is None or url == "":
+            raise Exception(
+                "You account information has gone stale! Please logout and login again!")
+
+        return url + relative_url
+
     def get_dataset_details_url(self, dataset_id: str, tab: str) -> str:
-        """ get the FASTGenomics url of dataset details tab """
+        """ get the url of dataset details tab """
         if tab == "":
-            return self._get_url(f"datasets/detail-{dataset_id}")
-        return self._get_url(f"datasets/detail-{dataset_id}#{tab}")
+            return self.get_platform_url(f"datasets/detail-{dataset_id}")
+        return self.get_platform_url(f"datasets/detail-{dataset_id}#{tab}")
+
+    def generic_get(self, url: str) -> Response:
+        """ Sends a GET request. Returns :class:`Response` object.
+
+        Parameters:
+          url : url (if it is relative, it will is relative to plattform)
+
+        Returns:
+          the response object of the request
+        """
+
+        if not is_absolute(url):
+            url = self.get_platform_url(url)
+
+        bearer_token = self._get_bearer_token()
+        headers = {
+            "Content-type": "application/json",
+            "Accept": "application/json",
+            "Authorization": f"Bearer {bearer_token}",
+            "User-Agent": user_agent,
+        }
+
+        response = self.session.get(
+            url, headers=headers
+        )
+        self._handle_response(response)
+        return response
+
+    def generic_put_json(self, url: str, json: str) -> Response:
+        """ Sends a JSON PUT request. Returns :class:`Response` object.
+
+        Parameters:
+          url : url (if it is relative, it will is relative to plattform)
+          json: a json string containing the data
+
+        Returns:
+          the response object of the request
+        """
+        if not is_absolute(url):
+            url = self.get_platform_url(url)
+
+        bearer_token = self._get_bearer_token()
+        headers = {
+            "Content-type": "application/json",
+            "Accept": "application/json",
+            "Authorization": f"Bearer {bearer_token}",
+            "User-Agent": user_agent,
+        }
+
+        response = self.session.put(
+            url, headers=headers, data=json.encode("utf-8")
+        )
+        self._handle_response(response)
+        return response
+
+    def generic_post_json(self, url: str, json: str) -> Response:
+        """ Sends a JSON POST request. Returns :class:`Response` object.
+
+        Parameters:
+          url : url (if it is relative, it will is relative to plattform)
+          json: a json string containing the data
+
+        Returns:
+          the response object of the request
+        """
+
+        if not is_absolute(url):
+            url = self.get_platform_url(url)
+
+        bearer_token = self._get_bearer_token()
+        headers = {
+            "Content-type": "application/json",
+            "Accept": "application/json",
+            "Authorization": f"Bearer {bearer_token}",
+            "User-Agent": user_agent,
+        }
+
+        response = self.session.post(
+            url, headers=headers, data=json.encode("utf-8")
+        )
+        self._handle_response(response)
+        return response
+
+    def generic_post_file(self, url: str, name: str, file_name: str, mime_type="application/octet-stream") -> Response:
+        """ Sends a file PUT multipart form request. Returns :class:`Response` object.
+
+        Parameters:
+          url          : url (if it is relative, it will is relative to plattform)
+          name         : the name of the file
+          file_name    : the path to file to be uploaded
+          mime_type    : the mime type
+
+        Returns:
+          the response object of the request
+        """
+
+        if not is_absolute(url):
+            url = self.get_platform_url(url)
+
+        bearer_token = self._get_bearer_token()
+        headers = {
+            "Authorization": f"Bearer {bearer_token}",
+            "User-Agent": user_agent
+        }
+
+        with open(file_name, 'rb') as f:
+            file = {'file': (name, f, mime_type)}
+            response = self.session.post(
+                url, headers=headers, files=file
+            )
+            self._handle_response(response)
+        return response
+
+    def generic_patch_json(self, url: str, json: str) -> Response:
+        """ Sends a JSON PATCH request. Returns :class:`Response` object.
+
+        Parameters:
+          url : url (if it is relative, it will is relative to plattform)
+          json: a json string containing the data
+
+        Returns:
+          the response object of the request
+        """
+        if not is_absolute(url):
+            url = self.get_platform_url(url)
+
+        bearer_token = self._get_bearer_token()
+        headers = {
+            "Content-type": "application/json",
+            "Accept": "application/json",
+            "Authorization": f"Bearer {bearer_token}",
+            "User-Agent": user_agent,
+        }
+
+        response = self.session.patch(
+            url, headers=headers, data=json.encode("utf-8")
+        )
+        self._handle_response(response)
+        return response
+
+    def generic_delete(self, url: str) -> Response:
+        """ Sends a DELETE request. Returns :class:`Response` object.
+
+        Parameters:
+          url : url (if it is relative, it will is relative to plattform)
+
+        Returns:
+          the response object of the request
+        """
+        if not is_absolute(url):
+            url = self.get_platform_url(url)
+
+        bearer_token = self._get_bearer_token()
+        headers = {
+            "Content-type": "application/json",
+            "Accept": "application/json",
+            "Authorization": f"Bearer {bearer_token}",
+            "User-Agent": user_agent,
+        }
+
+        response = self.session.delete(
+            url, headers=headers
+        )
+        self._handle_response(response)
+        return response
 
 
 class FASTGenomicsLargeFileStorageClient(FASTGenomicsClient):
@@ -899,7 +1121,7 @@ class FASTGenomicsLargeFileStorageClient(FASTGenomicsClient):
         try:
             bearer_token = self._get_bearer_token()
 
-            url = self._get_url("webclient/api/lfs")
+            url = self.get_backend_url("api/lfs")
 
             recipient_email = recipientInformation["userEmail"]
             logger.info(f"title            : {title}")
@@ -979,13 +1201,13 @@ class FASTGenomicsLargeFileStorageClient(FASTGenomicsClient):
 
         if state == LargeFileStorageState.ERROR:
             logger.info(f"setting status of storage {sai.storage_id} to ERROR")
-            url = self._get_url(
-                f"webclient/api/lfs/{sai.storage_id}/state/Error"
+            url = self.get_backend_url(
+                f"api/lfs/{sai.storage_id}/state/Error"
             )
         elif state == LargeFileStorageState.READY:
             logger.info(f"setting status of storage {sai.storage_id} to READY")
-            url = self._get_url(
-                f"webclient/api/lfs/{sai.storage_id}/state/Ready"
+            url = self.get_backend_url(
+                f"api/lfs/{sai.storage_id}/state/Ready"
             )
         else:
             raise exception(f"unknown state '{state}'")
@@ -1098,7 +1320,7 @@ class FASTGenomicsLargeFileStorageClient(FASTGenomicsClient):
 
         # create the dataset
         logger.info("creating dataset")
-        url = self._get_url("webclient/api/datasets")
+        url = self.get_backend_url("api/datasets")
 
         request_data = {"title": title, "typeId": dataset_type, "files": []}
 
@@ -1120,8 +1342,8 @@ class FASTGenomicsLargeFileStorageClient(FASTGenomicsClient):
 
         # attach the lfs
         logger.info("attaching lfs to dataset")
-        url = self._get_url(
-            f"webclient/api/lfs/{sai.storage_id}/dataset/{dataset_id}"
+        url = self.get_backend_url(
+            f"api/lfs/{sai.storage_id}/dataset/{dataset_id}"
         )
         headers = {
             "Content-type": "application/json",
@@ -1221,7 +1443,8 @@ class FASTGenomicsLargeFileStorageClient(FASTGenomicsClient):
                 print("Download the file by curl", file=f)
                 print("", file=f)
                 filename = path.basename(full_filename)
-                print(f'  curl -o "{filename}" "{signed_downloadurl}" ', file=f)
+                print(
+                    f'  curl -o "{filename}" "{signed_downloadurl}" ', file=f)
                 print("", file=f)
                 print(
                     f"  Extract files using 7-zip. Use this password to enable extraction",
@@ -1403,8 +1626,8 @@ class FASTGenomicsLargeFileStorageClient(FASTGenomicsClient):
         bearer_token = self._get_bearer_token()
 
         logging.info(f"getting download url for storage {storage_id}")
-        url = self._get_url(
-            f"webclient/api/lfs/{storage_id}/content/{access_token}"
+        url = self.get_backend_url(
+            f"api/lfs/{storage_id}/content/{access_token}"
         )
 
         headers = {
@@ -1482,7 +1705,7 @@ class FASTGenomicsDatasetClient(FASTGenomicsClient):
 
         # register the upload
         logger.info(f"register upload of file {filename} {filesize} bytes")
-        url = self._get_url("webclient/api/fileuploads")
+        url = self.get_backend_url("api/fileuploads")
 
         request_data = {"name": filename, "size": filesize}
         request_json = json.dumps(request_data)
@@ -1499,7 +1722,7 @@ class FASTGenomicsDatasetClient(FASTGenomicsClient):
 
         # attach to dataset
         logging.info(f"attaching upload {uploadId} to dataset {dataset_id}")
-        url = self._get_url(f"webclient/api/datasets/{dataset_id}/files")
+        url = self.get_backend_url(f"api/datasets/{dataset_id}/files")
 
         request_data = {"uploadId": uploadId, "type": filetype}
         request_json = json.dumps(request_data)
@@ -1531,8 +1754,8 @@ class FASTGenomicsDatasetClient(FASTGenomicsClient):
         with open(full_filename, "rb") as f:
             chunk = f.read(chunk_size)
             while chunk:
-                url = self._get_url(
-                    f"webclient/api/fileuploads/{uploadId}/chunk/{chunkId}"
+                url = self.get_backend_url(
+                    f"api/fileuploads/{uploadId}/chunk/{chunkId}"
                 )
                 headers = {
                     "Authorization": f"Bearer {bearer_token}",
@@ -1549,7 +1772,7 @@ class FASTGenomicsDatasetClient(FASTGenomicsClient):
         if progress_bar is not None:
             progress_bar.close()
 
-        url = self._get_url(f"/datasets/detail-{dataset_id}#Files")
+        url = self.get_platform_url(f"/datasets/detail-{dataset_id}#Files")
 
         r = {
             "dataset_id": dataset_id,
@@ -1567,7 +1790,7 @@ class FASTGenomicsDatasetClient(FASTGenomicsClient):
         bearer_token = self._get_bearer_token()
 
         logging.info(f"setting metadata of dataset {dataset_id}")
-        url = self._get_url(f"webclient/api/datasets/{dataset_id}/metadata")
+        url = self.get_backend_url(f"api/datasets/{dataset_id}/metadata")
 
         request_json = json.dumps(metadata)
         headers = {
@@ -1586,7 +1809,7 @@ class FASTGenomicsDatasetClient(FASTGenomicsClient):
         bearer_token = self._get_bearer_token()
 
         logging.info(f"deleting dataset {dataset_id}")
-        url = self._get_url(f"webclient/api/datasets/{dataset_id}")
+        url = self.get_backend_url(f"api/datasets/{dataset_id}")
 
         headers = {
             "Content-type": "application/json",
@@ -1602,7 +1825,7 @@ class FASTGenomicsDatasetClient(FASTGenomicsClient):
     ) -> requests.Response:
         logging.info(f"creating dataset {title}")
         bearer_token = self._get_bearer_token()
-        url = self._get_url("webclient/api/datasets")
+        url = self.get_backend_url("api/datasets")
         request_data = {"title": title, "typeId": dataset_type, "files": []}
         request_json = json.dumps(request_data)
         headers = {
@@ -1621,7 +1844,7 @@ class FASTGenomicsDatasetClient(FASTGenomicsClient):
         bearer_token = self._get_bearer_token()
 
         logging.info(f"getting dataset types")
-        url = self._get_url("webclient/api/datasettypes")
+        url = self.get_backend_url("api/datasettypes")
 
         headers = {
             "Content-type": "application/json",
@@ -1640,7 +1863,7 @@ class FASTGenomicsDatasetClient(FASTGenomicsClient):
         bearer_token = self._get_bearer_token()
 
         logging.info(f"getting dataset types")
-        url = self._get_url(f"webclient/api/datasettypes/{id}")
+        url = self.get_backend_url(f"api/datasettypes/{id}")
 
         headers = {
             "Content-type": "application/json",
@@ -1659,7 +1882,7 @@ class FASTGenomicsDatasetClient(FASTGenomicsClient):
         bearer_token = self._get_bearer_token()
 
         logging.info(f'getting editable fields for dataset type "{id}".')
-        url = self._get_url(f"webclient/api/datasettypes/{id}/editablefields")
+        url = self.get_backend_url(f"api/datasettypes/{id}/editablefields")
 
         headers = {
             "Content-type": "application/json",
